/* NMR spectrum calculator. Stripped version to use with Tcl/Tk interface */

/* Compile with a 32bit compiler */

/* This version works correctly with 32 bit extended Watcom.
*/

/* Usage: barespin infile outfile >messages
   (messages are written to standard output

   If first parameter begins with -L, the program calculates a Lorentz
   profile for a previously computed output.
*/

/* input file has the following format:
    line 1: number of spins (N)
    line 2: P or H for ppm or Hz, followed by spectrometer freq in Mhz
    line 3: chemical shifts
    line 4: J11 J12 J13 ...
    line 5: J23 J24 ...
            ...

   Output is a sorted list of Hz ppm intensity triples, if ppm is
   specified, Hz intens. duples, otherwise.
*/

#include <stdlib.h>
#include <math.h>
#include <float.h>
#include <string.h>
#include <stdio.h>
#include <assert.h>


#define  T9    ,43758
#define  LASTMASK ,0xff
#define  MAXHDIM  (MAXBLK)
#define  MAX   9     /* 32 bit version can do 9 spins */
#define  P2MAX 512
#define  MAXBLK 126

#define TRUE  1
#define FALSE 0
#define ERROR  -1

#define ROUND      1e-9    /* consolidate lines closer than this */

#define MININT 5e-5     /* forget transitions weaker than this */
#define ACC    1e-9     /* accuracy parameter. seek this fraction of
                           smallest j as eigenvalue accuracy */


struct transition {  double de;
                     double intens;
                  };

void errmsg(char *f, char *s);
int read_file(char *n);
int qgets(char *b, int size, FILE *f);
void nmrcalc( int nspin, double *freq, double j[MAX][MAX], struct transition *t);
int bitcount(int i, int n);
int ymeig(int n, double *a, double *e, double eps, int v);
void tssort(struct transition *data, int n);
void lorentzcalc (int argc, char **argv);


/* global variables */
   int ppm, nspin;
   double sf, inpfreq[MAX], j[MAX][MAX];


main(int argc, char **argv)
{
   int i,  kk, nt, lorentz, recalc;
   static int ntran[MAX+1] = { 0, 1, 4, 15, 56, 210, 792, 3003, 11440 T9};
   double freq[MAX];
   struct transition *trans;
   FILE *output;


/* explanation of variables:
   ntran[n] contains number of transitions in general n-spin system.
   freq and j are user input frequencies and coupling constants. (Only
      upper triangle of j is actually used .)
   trans will point to  alloced space for transition freqs and intensities.
*/


   if(argc < 3)
   {  errmsg("Must have input and output file arguments", "");
      exit(1);
   }

   if (*argv[1] == '-' && *(argv[1] + 1) == 'L') lorentzcalc(argc, argv);

   ppm = ERROR;
   ppm = read_file(argv[1]);
   if (ppm == ERROR)
   {  errmsg("Can't open", argv[1]);
      exit(1);
   }

/* assert(ppm == TRUE || ppm == FALSE); */

   if (NULL == (trans = (struct transition *)calloc(ntran[nspin],
               sizeof(struct transition)))) errmsg("main", "Out of memory");


   lorentz = 0;
   recalc = 0;

   if (ppm) for (i=0;  i<nspin;  i++) freq[i] = sf * inpfreq[i];
   else for (i=0;  i<nspin;  i++) freq[i] = inpfreq[i];

    nmrcalc(nspin, freq, j, trans);

    for (i=0, nt=0;  i<ntran[nspin];  i++)
    {  if (trans[i].intens < MININT) continue;  /* squeeze out very weak */
       else
       {   trans[nt] = trans[i];
           nt++;
       }
    }
   tssort(trans, nt);                           /* sort transitions */

   for (i=0, kk=0;  i<nt;  i++)                 /* consolidate duplicates */
   {  if (kk>0 && fabs(trans[i].de - trans[kk-1].de) < ROUND)
      {   trans[kk-1].intens += trans[i].intens;
          continue;
      }
      else
      {   trans[kk] = trans[i];
          kk++;
      }
   }
   nt = kk;
   if(ppm) for (i=0;  i<nt;  i++) trans[i].de /= sf;  /* plot will be in ppm */

   output = fopen(argv[2], "wt");
   if (output == NULL)
   {  errmsg("Can't open output file ", argv[2]);
      exit(1);
   }
   if(ppm)                          /* put out Hz, ppm, intensity */
   {  for(i=0;  i<nt;  i++)
         fprintf(output, "%11.5f %12.7f  %7.4f\n", sf*trans[i].de, trans[i].de, trans[i].intens);
   }
   else                             /* Hz and intensity only */
   {  for(i=0;  i<nt;  i++)
         fprintf(output, "%11.5f  %7.4f\n", trans[i].de, trans[i].intens);
   }

   fclose(output);
   exit(0);
}




void tssort(struct transition *data, int n)
                 /* Shell sort, specific to struct transition */
{
    unsigned j, k, m, datastep;
    int i, mptrstep;
    struct transition  *p1,  *p2, temp;

    if (n <= 1) return;

    for (m=1;  m<=n;  m<<=1) ;  /* calculate the Hibbert step length */
    m >>= 1;
    m--;

    while (m>0)
    {   k = n - m - 1;
        mptrstep = m;
        datastep = 0;
        for (j=0;  j<=k;  j++, datastep ++)
        {   i = j;
            p1 = data + datastep;
            p2 = p1 + mptrstep;
            while ( i >= 0  && p1->de > p2->de)
            {   temp = *p1;
                *p1 = *p2;
                *p2 = temp;
                i -= m;
                p1 -= mptrstep;
                p2 -= mptrstep;
            }
        }
        m >>= 1;
    }
}



void nmrcalc( int nspin, double *freq, double j[MAX][MAX],
                                       struct transition *t)
{
   static int nlast = 0;      /* holds previous value of nspin */
   static int bnc[MAX+1];     /* binomial coefficients */
   static int basis[P2MAX];   /* coded basis functions */
   static int bindex[P2MAX];  /* element i is locn of fn i in basis[] */
   static int start[MAX+1];   /* indexes to successive Fz blocks in basis[] */
   static int bits[MAX];      /* powers of 2 */

/* elements of basis code the spin product basis functions in bottom nspin
   bits. 1 bit for alpha, 0 bit for beta */

   char buf[81];
   int r, b;
   register int a, q;
   int i, k, m, f, n, bi, bk, n2, br, s2;
   int mask[MAX] = {0x1fe, 0x1fd, 0x1fb, 0x1f7, 0x1ef, 0x1df, 0x1bf, 0x17f LASTMASK};
   double /*far*/ *hspace[MAX+1];
   double /*far*/ *h[MAXBLK], /*far*/ *h2[MAXBLK], *e, *e2, ti, eps, hrk, hii;

   if (nspin != nlast)        /* set up bnc, basis, etc if nspin changed */
   {  bnc[0] = 1;
      start[0] = 0;
      for(i=1, k=nspin, m=1;  i<=MAX;  i++, k--, m++)
      {  bnc[i] = (bnc[i-1] * k) / m;
         start[i] = start[i-1] + bnc[i-1];
      }
      m = 1 << nspin;
      for (i=0;  i<m;  i++)
      {  k = bitcount(i, nspin);
         bindex[i] = start[k];
         basis[(start[k]++)] = i;
      }
      start[0] = 0;
      for (i=1;  i<=MAX;  i++) start[i] = start[i-1] + bnc[i-1];

      for (i=0;  i<nspin;  i++) bits[i] = 1 << i;

      nlast = nspin;
   }

   eps = 1e100;
   for (i=0;  i<nspin;  i++)
   {  for (k=i+1;  k<nspin;  k++)
      {  if (j[i][k] == 0) continue;
         if (fabs(j[i][k]) < eps) eps = fabs(j[i][k]);
      }
   }
   if (eps == 0) eps = 1;
   eps *= ACC;             /* smallest j defines accuracy sought in energies */


   for (f=0;  f<= nspin;  f++)  /* loop over values of Fz, allocating, setting up
                                 and solving each block of the Hamiltonian. For
                                 all but first block, calculate transitions.
                                 Deallocate old blocks at bottom */
   {  n = bnc[f];
      if (NULL == (hspace[f] = (double *)malloc((unsigned)n*n*sizeof(double))))
             errmsg("nmrcalc", "Out of memory");

      if (f != 0)
      {  for (i=0;  i<bnc[f-1];  i++) h2[i] = h[i];  /*save prev row ptrs */
         e2 = e;                                     /* and energy */

      }
      if (NULL == (e = (double *)malloc(n*sizeof(double))))
             errmsg("nmrcalc", "Out of memory");
                                                    /* space for energies */

      for (i=0;  i<n;  i++)      /* set pointers to rows of matrix block */
         h[i] = hspace[f] + i * n;

      for (i=0;  i<n;  i++)
      {  for (k=i;  k<n;  k++)
         {  h[i][k] = 0;         /* zero the upper triangle */
         }
         bi = basis[start[f] + i];

         hii = 0;
         for (a=0;  a<nspin;  a++)
         {  if (bi & bits[a]) hii -= freq[a];
            else hii += freq[a];
         }
         hii *= .5;           /* done shift contributions to diagonal */

         for (a = 0;  a < nspin;  a++)     /* loop through couplings */
         {  for (b = a+1;  b < nspin;  b++)
            {  q = bi & bits[a];
               r = bi & bits[b];
               if ((q&&r) || (!(q||r))) hii += .25 * j[a][b];
               else hii -= .25 * j[a][b];
            }
         }                   /* done J contributions to diagonal */
         h[i][i] = hii;

         for (k=i+1;  k<n;  k++)    /* do off-diagonal terms */
         {  bk = basis[start[f] + k];
            q = bk ^ bi;      /* there is an off-diag iff basis fns differ
                                 in exactly 2 positions. The element is .5
                                 times the j for these 2 posns */
            if (bitcount(q, n) != 2) continue;
            for (m=0;  m<n;  m++)
            {  if (q & 1)
               {  a = m++;
                  q >>= 1;
                  break;
               }
               q >>= 1;
            }
            for ( ;  m<n;  m++)
            {  if (q & 1)
               {  b = m;
                  break;
               }
               q >>= 1;
            }
            h[i][k] = .5 * j[a][b];
         }                             /* off-diagonals done */
         for (k=i+1;  k<n;  k++) h[k][i] = h[i][k];

      }                                /* matrix block is complete */


      q = ymeig(n, h[0], e, eps, 1);   /* solve the current block */

      if (q)
      {   sprintf(buf,"error %d", q);
          errmsg("ymeig", buf);
      }

/* calculate transition energies and intensities, involving this block and
   the previous one */

      if (f == 0) continue;            /* nothing on 1st block */


      n2 = bnc[f-1];
      for (i=0;  i<n2; i++)            /* i loops states in prev block */
      {  for (k=0;  k<n;  k++)         /* k loops present block */
         {  t->de = e2[i] - e[k];      /* energy done */
            s2 = start[f-1];
            ti = 0;
            for (r=0;  r<n;  r++)      /* r loops basis fns in present state */
            {  br = basis[start[f] + r];
               hrk = h[r][k];
               for (q=0;  q<nspin;  q++)   /* loop over spins */
               {   a = br & mask[q];       /* if spin q is alpha, a is the */
                   if (a != br)            /* basis fn of next lower f, to */
                   {  a = bindex[a];       /* which transition is possible */
                      ti += h2[a-s2][i] * hrk;  /* lookup tells which fn. */
                   }
               }
            }
            t->intens = ti * ti;
            t++;
         }
      }

      free(hspace[f-1]);            /* free dead blocks */
      free(e2);

      if (f == nspin)
      {  free(hspace[f]);
         free(e);
      }

   }  /* end of loop over blocks */
   return;
}



bitcount(int i,  int n)    /* count no of 1's in bottom n bits of i */
{
   int k, m;

   k = 0;
   for (m=0;  m<n;  m++)
   {  if (i & 1) k++;
      i >>= 1;
   }
   return(k);
}



void ymeisb(long double p, long double q, double *c, double *s, double *n);
/* auxiliarry for ymeig */


int ymeig(n, aa, d, eps, vector)
   int n, vector;
   double *aa, *d, eps;
{
/* Transcription of Fortran routine of same name. Stewart's algorithm
   (CACM 13, 370 ( 1970), errata ibid 750) for eigenvalues, and
   optionally eigenvectors, of a real symmetric matrix.

   Arguments:
   n     dimension of matrix (must be <= 90 in 16 bit code)
   aa    input matrix. This is stored in full in a memory block of exactly
         the right size. (eg mallocked). C storage order; i.e. row-wise.
   d     output vector of length n for eigenvalues
   eps   absolute required convergence in eigenvalues
   vector if != 0, this will cause aa to be overwritten with eigenvectors

   Return value is error code, as follows:
      0: OK
     -1: Invalid n (<1 or > MAXHDIM)
     -2: Unable to malloc required workspace
    n>0: Convergence failed. n = number of initial unreliable results.
*/


   double **a, *e,  k1, ninf, norm, qmax, s2, con, temp, test;
   double cb, cc, cd, c, s, r;
   double *aj, *ai, *aLL;
   long double sum, sum1, k, k2, p, q;
   int shft, i, j, l, L1, LL, LL1, nu, num1, nl;

   double k0 = 0.2;              /* initial origin shift */
   int itlim = 50;               /* maximum allowed iterations to converge */


   if (n <= 0 || n > MAXHDIM) return(-1);

   if (n == 1)                   /* degenerate case */
   {  d[0] = aa[0];
      if (vector) aa[0] = 1;
      return(0);
   }

                                 /* allocate pointers to rows of aa */

   if (NULL == (a = (double **)malloc(n * sizeof(double *)))) return(-2);
   for (i=0;  i < n;  i++) a[i] = aa + i * n;

                                 /* allocate workspace vector */

   if (NULL == (e = (double *)malloc(n * sizeof(double))))
   {  free(a);
      return(-2);
   }


                                 /* real calculation for non-degenerate cases */

   if (n != 2)
   {  for (l=0;  l < n-2;  l++)  /* MAIN LOOP */
      {  L1 = l + 1;
         d[l] = a[l][l];
         qmax = 0;
         for (i=L1;  i < n;  i++)
         {  q = fabs(a[i][l]);
            if (qmax < q) qmax = q;
         }
         if (qmax == 0)
         {  e[l] = 0;
            a[l][l] = 1;
            continue;            /* to bottom of main loop */
         }
         sum = 0;
         for (i = L1;  i < n;  i++)
         {  ai = a[i];
            ai[l] /= qmax;
            sum += (ai[l] * ai[l]);
         }
/* assert(sum >= 0); */
         s2 = sqrt(sum);
         if (a[L1][l] < 0) s2 = -s2;
         e[l] = -s2 * qmax;
         a[L1][l] += s2;
         a[l][l] = s2 * a[L1][l];
         sum1 = 0;
         for (i=L1;  i<n;  i++)
         {  sum = 0;
            ai = a[i];
            for (j = L1;  j <= i;  j++) sum += (ai[j]) * (a[j][l]);
            if (i != n-1) for (j = i+1;  j < n;  j++)
                          {   aj = a[j];
                              sum += aj[l] * aj[i];
                          }
            e[i] = sum / a[l][l];
            sum1 += ai[l] * e[i];    /* 50 */
         }
         con = .5 * sum1 / a[l][l];
         for (i = L1;  i < n;  i++)
         {  e[i] -= con * a[i][l];
            ai = a[i];
            for (j = L1;  j <= i;  j++)
                     ai[j] -= (ai[l] * e[j]  +  a[j][l] * e[i]);
         }
      }                          /* 70 */
   }                             /* end of if ( n != 2) */

   d[n-2] = a[n-2][n-2];
   d[n-1] = a[n-1][n-1];
   e[n-2] = a[n-1][n-2];

   if (vector)
   {  a[n-2][n-2] = a[n-1][n-1] = 1;
      a[n-2][n-1] = a[n-1][n-2] = 0;

      if (n != 2)
      {  for (LL = n-3;  LL >= 0;  LL--)
         {  LL1 = LL + 1;
            aLL = a[LL];
            for (i = LL1;  i < n;  i++)
            {  sum = 0;
               for (j = LL1;  j < n;  j++)  sum += a[j][LL] * a[j][i];
               aLL[i] = sum / aLL[LL];
            }                    /* 140 */

            for (i = LL1;  i < n;  i++)
            {  ai = a[i];
               for (j = LL1;  j < n;  j++) ai[j] -= ai[LL] * aLL[j];
            }
            for (i = LL1;  i < n;  i++) a[i][LL] = aLL[i] = 0;
            aLL[LL] = 1;
         }                       /* 170 */
      }
   }

   ninf = (fabs(d[0]) + fabs(e[0]) > fabs(d[n-1]) + fabs(e[n-2])) ?
               (fabs(d[0]) + fabs(e[0])) : ( fabs(d[n-1]) + fabs(e[n-2]));

   if (n != 2) for (i = 1;  i < n-1;  i++)
               {  q = fabs(d[i]) + fabs(e[i]) + fabs(e[i-1]);
                  if (ninf < q) ninf = q;
               }

   nu = n - 1;
   num1 = nu - 1;
   shft = 0;
   k1 = k0;
   test = ninf * eps;
   e[n-1] = 0;

   while(1)             /* iterate to convergence or failure */
   {  for (nl = num1;  nl >= 0;  nl--)
      {  if (fabs(e[nl]) <= test ) break;
      }
      if (nl >= 0)
      {  e[nl++] = 0;            /* 230 */
         if (nl == nu)
         {  if (num1 == 0)
            {  free(a);
               free(e);
               return(0);
            }
            nu = num1;
            num1 = nu - 1;
            continue;            /* rerun while(1) loop */
         }
      }
      nl = 0;                    /* 240 */
      e[nu] += 1;
      if (e[nu] > itlim)
      {  free(a);
         free(e);
         return(nu);
      }

      cb = (d[num1] - d[nu]) / 2;   /* 250 */
      qmax = (fabs(cb) > fabs(e[num1])) ? fabs(cb) : fabs(e[num1]);
/* assert(qmax != 0); */
      cb /= qmax;
      cc = e[num1] / qmax;
      cc *= cc;
/* assert(cb * cb + cc >= 0); */
      cd = sqrt(cb * cb + cc);
      if(cb != 0) cd = fabs(cd) * ((cb > 0) ? 1 : -1);
/* assert(cb + cd != 0); */
      k2 = d[nu] - qmax * cc / (cb + cd);

      if (shft) k = k2;
      else if (fabs(k2 - k1) < .5 * fabs(k2))
      {  shft = 1;
         k = k2;
      }
      else
      {  k1 = k2;
         k = k0;
      }

      ymeisb(d[nl] - k, e[nl], &c, &s, &norm);     /* 300 */

      for (i = nl;  i <= num1;  i++)
      {  if(vector)
         {  for (j=0;  j < n;  j++)
            {  aj = a[j];
               temp = c * aj[i] + s * aj[i+1];
               aj[i+1] = -s * aj[i] + c * aj[i+1];
               aj[i] = temp;
            }
         }
         d[i] = c * d[i] + s * e[i];
         temp = c * e[i] + s * d[i+1];
         d[i+1] = -s * e[i] + c * d[i+1];
         e[i] = -s * k;
         d[i] = c * d[i] + s * temp;
         if (i == num1) continue;
         if (fabs(s) <= fabs(c))
         {  assert(c != 0);
            r = s / c;
            d[i+1] = -s * e[i] + c * d[i+1];
            e[i+1] *= c;
            ymeisb(d[i+1] - k, e[i+1], &c, &s, &norm);
            e[i] = r * norm;
         }
         else
         {  p = c * e[i] + s * d[i+1];
            q = s * e[i+1];
/* assert(s != 0); */
            d[i+1] = k + c * p / s;
            e[i+1] *= c;
            ymeisb(p, q, &c, &s, e+i);
         }
      }                                /* 380 */

      temp = c * e[num1] + s * d[nu];
      d[nu] = -s * e[num1] + c * d[nu];
      e[num1] = temp;
   }                                   /* end of while(1) iteration */
}


void ymeisb(long double p, long double q, double *c, double *s, double *n)
{
   if (fabs(q) >= fabs(p))    /* original has > but that doesnt make sense */
   {  if (q == 0)
      {  *c = 1;
         (*s) = (*n) = 0;
      }
      else
      {  assert(q != 0);
         *n = fabs(q) * sqrt(1 + p * p / (q * q));
/* assert(*n != 0); */
         (*c) = p / (*n);
         (*s) = q / (*n);
      }
   }
   else
   {  if(p == 0) (*n) = q;
/* limiting form of following statement, which crashes for p = 0. Should
   this be possible? */
      else (*n) = fabs(p) * sqrt(1 + q * q / (p * p));

/* assert((*n) != 0); */
      (*c) = p / (*n);
      (*s) = q / (*n);
   }
}



int read_file(char *n)
{  char b[81], *p;
   FILE *input;
   int i, k, ppm, len;

   input = fopen(n, "rt");
   if (input == NULL)
   {  errmsg("Can't open", n);
      return(ERROR);
   }
   if (EOF == qgets(b, 81, input)) return(ERROR);
   nspin = atoi(b);
   if (nspin < 2 || nspin > MAX) goto bad_input;

   if (EOF == qgets(b, 81, input)) return(ERROR);
   len = strlen(b);
   for(i=0;  i<len;  i++)
   {    if (b[i] >= 'a' && b[i] <= 'z') b[i] += ('A' - 'a');
   }
/* write out 'strupr()' ... beware weird character encodings */

   p = strpbrk(b, "PH");
   if (p == NULL) goto bad_input;
   if (*p == 'P') ppm = TRUE;
   else ppm = FALSE;
   if (ppm) sf = atof(p + 1);

   if (EOF == qgets(b, 81, input)) return(ERROR);
   p = b;
   for (i=0;  i<nspin;  i++)
   {  if (*p == 0) goto bad_input;
      inpfreq[i] = strtod(p, &p);
   }
   for (i=0;  i<nspin - 1;  i++)
   {  if (EOF == qgets(b, 81, input)) return (ERROR);
      p = b;
      for (k=i+1;  k<nspin;  k++)
      {  if (*p == 0) goto bad_input;
         j[i][k] = strtod(p, &p);
      }
   }
   return(ppm);

bad_input:
   errmsg("Bad input:", b);
   return(ERROR);
}



int qgets(char *b, int size, FILE *f)  /* like fgets but eliminates comments */
{  char *p;                            /* starting with #. returns EOF on */
   int i, n;                           /* EOF, TRUE otherwise */

   do
   {  p = fgets(b, size, f);
      if (p == NULL) return(EOF);

      p = strchr(b, '#');
      if (p != NULL) *p = 0;
      n = strlen(b);
      for (i=n-1;  i>=0;  i--)
      {  if (b[i] == ' ' || b[i] == '\n' ) b[i] = 0;
                                 /* strip trailing blanks or newlines */
         else break;
      }
   } while (strlen(b) == 0);

   return(TRUE);
}

void errmsg(char *s1, char *s2)
{  printf("%s %s\n", s1, s2);
}



void lorentzcalc (int argc, char **argv)
{
/* calculate lorentzians. argv[1] must be of form -LPxxx or -LH where xxx is
   spectrometer freq in Mhz. P -> ppm calc, H-> Hz. argv[2] is h-p-i or
   h-i data file produced by previous calculation, argv[3] is output
   file name, argv[4] and [5] are left and right freq or ppm (based on
   argv[1], argv [6] is number of data points to calculate, argv[7] is
   f.w.h.m. in Hz. argv[8] is height in pels, argv[9] is leftmost x point
   in pels. Output is lines with x,y pairs in pels.
*/

   double left, right, fwhm, sfmhz, del, b, n1f, maxi, reli;
   FILE *in, *out;
   int points, nt, i, j, k, n, n1, n2, n3, first, last, space, jleft;
   char ppm, bb[81], *ep;
   float *x, *intens, *result;
   float z, zz, height, baseline;

   if(argc < 11)
   {  errmsg("Not enough arguments for Lorentz calculation", "");
      exit(1);
   }
   if(argv[1][2] == 'P') ppm = 1;
   else if(argv[1][2] == 'H') ppm = 0;
   else
   {  errmsg("Bad arg1 for lorentz calc", "");
      exit(1);
   }
   if(ppm)
   {  sfmhz = atof(argv[1] + 3);
      if(sfmhz <= 0)
      {  errmsg("Bad spec freq", "");
         exit(1);
      }
   }

   in = fopen(argv[2], "rt");
   if(in == NULL)
   {  errmsg("(L) Can't open", argv[2]);
      exit(1);
   }
   out = fopen(argv[3], "wt");
   if(out == NULL)
   {  errmsg("(L) Can't open", argv[3]);
      exit(1);
   }

   left = atof(argv[4]);
   right = atof(argv[5]);
   points = atoi(argv[6]);
   fwhm = atof(argv[7]);
   height = atof(argv[8]);
   baseline = atof(argv[9]);
   jleft = atoi(argv[10]);
   
   if(left <= right)
   {  errmsg("Bad left/right", "");
      exit(1);
   }
   if(points <=2)
   {  errmsg("Not enough points", "");
      exit(1);
   }
   if(fwhm < 0)
   {  errmsg("Bad width", "");
      exit(1);
   }

   x = malloc(2000 * sizeof(float));
   intens = malloc(2000 * sizeof(float));
   space = 2000;
   result = malloc(points * sizeof(float));

   if(x == NULL || intens == NULL || result == NULL)
   {  errmsg("Out of memory", "");
      exit(1);
   }

   nt = 0;
   while(1)
   {  if(NULL == fgets(bb, 81, in)) break;
      x[nt] = strtod(bb, &ep);
      if(ppm == 1) b = strtod(ep, &ep);   /* discard */
      intens[nt] = strtod(ep, &ep);
      nt++;
      if(nt >= space - 1)
      { space += 2000;
        x = realloc(x, space * sizeof(float));
        intens = realloc(intens, space * sizeof(float));
        if(x == NULL || intens == NULL)
        {   errmsg("out of space", "");
            exit(1);
        }
      }
/* read data to x, intens;  update nt */
   }

   if(ppm == 1)
   {  right *= sfmhz;
      left *= sfmhz;
   }


   for(i=0;  i<points;  i++) result[i] = 0;

   b = 4/(fwhm * fwhm);
   del = fabs(left - right) / points;
   n1f = 32 * fwhm / del;

   n1 = n1f;
   n2 = n1 / 3.0;
   n3 = n2 / 3.0;

   for (first=0;  first<nt;  first++)
               if (x[first] >= right) break;     /* NMR! Backwards! */
   if (first >= nt || left < x[first])
   {   for(i=0;  i<points;  i++) result[i] = 1.0;
       goto data_out;   /* do nothing if no transitions in range */
   }

   for (last = first;  last <nt;  last++)
                if (x[last] > left ) break;
   last--;

   maxi = 0;
   for (i=first;  i<=last;  i++)
       if (intens[i] > maxi) maxi = intens[i];

   for (i = first;  i <= last;  i++)
   {   reli = intens[i] / maxi;
       if ( reli < .0005) continue;        /* We ignore lines so weak that they */
       else if (reli < .01) n = n3;       /* can't contribute .05% intensity */
       else if (reli < .1)  n = n2;       /* lorentzians are truncated at same */
       else n = n1;                       /* accuracy level */

       j = .5 + (left - x[i]) / del;
       if (j >= points) j = points - 1;

       zz = z = x[i] - (left - j * del);
       for (k = j;  (k-j) < n && k < points;  k++)
       {   result[k] += intens[i] / ( 1 + b * z * z);
           z += del;
       }
       z = zz - del;
       for (k = j-1;  (j-k) < n && k >= 0;  k--)
       {   result[k] += intens[i] / (1 + b * z * z);
           z -= del;
       }
   }


   maxi = 0;
   for (i=0;  i<points;  i++) if (maxi < result[i]) maxi = result[i];
   if(maxi == 0) maxi= 1;
   for(i=0;  i<points;  i++)
   {   result[i] = baseline - height * result[i] / maxi;
       if (result[i] < 0) result[i] = 0;
   }

   fclose(in);

/* write the result file */
data_out:
   for(i=0;  i<points;  i++)
   {    fprintf(out, "%d %.0f\n", jleft, result[i]);
        ++jleft;
   }
   fclose(out);

   exit(0);
}

